/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

public class ReadRandom {
	private static final int DEFAULT_BUFSIZE = 4096;

	private RandomAccessFile raf;
	private byte inbuf[];
	private long startpos = -1;
	private long endpos = -1;
	private int bufsize;

	public ReadRandom(File file, int b) throws FileNotFoundException {
		raf = new RandomAccessFile(file, "r");
		bufsize = b;
		inbuf = new byte[bufsize];
	}
	
	public ReadRandom(File file) throws FileNotFoundException {
		this(file, DEFAULT_BUFSIZE);
	}
	
	public ReadRandom(String name) throws FileNotFoundException {
		this(name, DEFAULT_BUFSIZE);
	}

	public ReadRandom(String name, int b) throws FileNotFoundException {
		raf = new RandomAccessFile(name, "r");
		bufsize = b;
		inbuf = new byte[bufsize];
	}

	public int read(long pos) {
		if (pos < startpos || pos > endpos) {
			long blockstart = (pos / bufsize) * bufsize;
			int n;
			try {
				raf.seek(blockstart);
				n = raf.read(inbuf);
			} catch (IOException e) {
				return -1;
			}
			startpos = blockstart;
			endpos = blockstart + n - 1;
			if (pos < startpos || pos > endpos)
				return -1;
		}

		return inbuf[(int) (pos - startpos)] & 0xffff;
	}
	/*
	public String readLine(long pos) {
		StringBuffer sb = new StringBuffer();
		int c;
		
		long p = pos;
		
		while((c = read(p)) != -1) {
			if ((char)c == '\r' || (char)c == '\n')
				break;
			sb.append((char)c);
			p++;
		}
		
		return sb.toString();
	}
*/
	public void close() throws IOException {
		raf.close();
	}

}
