/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.plugins.charts;

import java.util.ArrayList;

import org.jfree.chart.JFreeChart;
import org.jfree.data.general.DefaultPieDataset;

import pt.unl.fct.di.jtraceview.trace.Event;
import pt.unl.fct.di.jtraceview.trace.TraceFileReader;
import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxReadEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxWriteEvent;

public class AbortChart implements MultipleChartType {

	private TraceFileReader reader;
	private ArrayList<JFreeChart> charts;
	private long user;
	private long commit;
	private long other;
	private long total;
	private long read;
	private long write;
	
	public AbortChart() {
		user = 0;
		commit = 0;
		other = 0;
		total = 0;
		read = 0;
		write = 0;
	}
	
	
	
	@Override
	public ArrayList<JFreeChart> getCharts() {
		return charts;
	}


	@Override
	public void init(TraceFileReader traceFile) {
		reader = traceFile;
		reader.reset();
		Event e;
		
		Event[] threadId = new Event[24];
		
		while((e = reader.getNextEvent()) != null) {
			if (e instanceof TxAbortEvent) {
				TxAbortEvent ev = (TxAbortEvent)e;
				
				if (ev.tid != -1) {
					if (ev.type == TxAbortEvent.USER) {
						user++;
					}
					else if (ev.type == TxAbortEvent.COMMIT) {
						commit++;
					}
					else if (ev.type == TxAbortEvent.OTHER) {
						Event ev2 = threadId[e.threadId];
						
						if (ev2 instanceof TxReadEvent) {
							read++;
						}
						else if (ev2 instanceof TxWriteEvent) {
							write++;
						}
						else {
							assert(false);
						}

						other++;
					}	
				}
				total++;
			} 
			
			threadId[e.threadId] = e;
		}

		DefaultPieDataset data = new DefaultPieDataset();
		data.setValue("User", (user/(double)total)*100);
		data.setValue("Commit", (commit/(double)total)*100);
		//data.setValue("Read/Write", (other/(double)total)*100);
		data.setValue("Read", (read/(double)total)*100);
		data.setValue("Write", (write/(double)total)*100);
		
		JFreeChart chart = ChartCreator.createPieChart("Abort Percentages", data);
		charts = new ArrayList<JFreeChart>();
		charts.add(chart);
	}

}
