/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;

import pt.unl.fct.di.jtraceview.trace.Event;
import pt.unl.fct.di.jtraceview.trace.TraceFileReader;

import pt.unl.fct.di.jtraceview.trace.events.TxCommitEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxReadEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxStartEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxWriteEvent;

public class GenerateDot {

	private TraceFileReader reader; // reader of the trace file

	// Types of dependencies
	private static final short WRITE_READ  = 0;
	private static final short WRITE_WRITE = 1;
	private static final short READ_WRITE  = 2;

	// Logs the events on this memCel
	// including the last transaction that wrote and the transactions that read the value
	// HashMap <MemCel, LogEvent>
	private HashMap<String, MemCellAccess> memCels;

	// Thread x is running transactionY
	// HashMap <Thread, Transaction_number>
	private HashMap<Integer,ThreadInfo> thread_trans;

	// Connections between transactions
	private ArrayList<Edge> edges;

	// Transactions count
	private int t_count;

	private int testPos = 0;

	public GenerateDot(TraceFileReader tfr) {
		reader = tfr;
	}

	public ArrayList<Edge> processFile() {
		reader.reset();
		memCels 	 = new HashMap<String,MemCellAccess>();
		thread_trans = new HashMap<Integer,ThreadInfo>();
		edges		 = new ArrayList<Edge>();
		t_count 	 = 0;
		
		collectInfo();
		createEdges();
		

//		Collections.sort( edges, new Comparator<Edge>()
//        {
//        public int compare( Edge a, Edge b )
//           {
//           return ( a.vertA - b.vertA );
//           }
//        } );
		
		return edges;
	}


	// WRITE SKEW do paper
	public Event testw() {

		Event e = null;
		// Hws : R1 (x = 20) R2 (x = 20) R1 (y = 80) R2 (y = 80) R1 (x = 20)
		//		W1 (x = −10) C1 R2 (y = 80) W2 (y = −10) C2
		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxReadEvent(testPos, 1, 0, "X"); break;

		case 2: e = new TxStartEvent(testPos, 2, 0); break;
		case 3: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		
		case 4: e = new TxReadEvent(testPos, 1, 0, "Y"); break;
		case 5: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		
		case 6: e = new TxReadEvent(testPos, 1, 0, "X"); break;
		case 7: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 8: e = new TxCommitEvent(testPos, 1, 0); break;
		
		case 9: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		case 10: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 11: e = new TxCommitEvent(testPos, 2, 0); break;

		default: e = null; break;
		}
		testPos++;
		return e;
	}
	// SI READ ONLY do paper
	public Event testSI() {

		Event e = null;

		// H : R2(x = 0), R2(y = 0), R1(y = 0), W1(y = 20), C1
		//			R3(x = 0), R3(y = 20), C3, W2(x = −11), C2
		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 2, 0); break;
		case 1: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		case 2: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		
		case 3: e = new TxStartEvent(testPos, 1, 0); break;
		case 4: e = new TxReadEvent(testPos, 1, 0, "Y"); break;
		case 5: e = new TxWriteEvent(testPos, 1, 0, "Y"); break;
		case 6: e = new TxCommitEvent(testPos, 1, 0); break;
		
		case 7: e = new TxStartEvent(testPos, 3, 0); break;
		case 8: e = new TxReadEvent(testPos, 3, 0, "X"); break;
		case 9: e = new TxReadEvent(testPos, 3, 0, "Y"); break;
		case 10: e = new TxCommitEvent(testPos, 3, 0); break;

		case 11: e = new TxWriteEvent(testPos, 2, 0, "X"); break;
		case 12: e = new TxCommitEvent(testPos, 2, 0); break;
		default: e = null; break;
		}
		testPos++;
		return e;
	}
	// PRIMEIRO do paper
	public Event test1() {
		Event e = null;

		//H3 : W1 (X) W1 (Y ) W1 (Z) C1 W3 (X) R2 (X) W2 (Y ) C2 R3 (Y ) C3
		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 2: e = new TxWriteEvent(testPos, 1, 0, "Y"); break;
		case 3: e = new TxWriteEvent(testPos, 1, 0, "Z"); break;
		case 4: e = new TxCommitEvent(testPos, 1, 0); break;
		
		case 5: e = new TxStartEvent(testPos, 3, 0); break;
		case 6: e = new TxWriteEvent(testPos, 3, 0, "X"); break;
		
		case 7: e = new TxStartEvent(testPos, 2, 0); break;
		case 8: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		case 9: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 10: e = new TxCommitEvent(testPos, 2, 0); break;
		
		case 11: e = new TxReadEvent(testPos, 3, 0, "Y"); break;
		case 12: e = new TxCommitEvent(testPos, 3, 0); break;
		default: e = null; break;
		}
		testPos++;
		return e;
	}

	
	// SI-Detection-part1.pdf page 56
	public Event testSIDetection() {
		Event e = null;
		// H: W1(X) W1(Y) C1 W3(X) R2(X) W2(Y) C2 R3(Y) C3
		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 2: e = new TxWriteEvent(testPos, 1, 0, "Y"); break;
		case 3: e = new TxCommitEvent(testPos, 1, 0); break;
		
		case 4: e = new TxStartEvent(testPos, 3, 0); break;
		case 5: e = new TxWriteEvent(testPos, 3, 0, "X"); break;
		
		case 6: e = new TxStartEvent(testPos, 2, 0); break;
		case 7: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		case 8: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 9: e = new TxCommitEvent(testPos, 2, 0); break;
		
		case 10: e = new TxReadEvent(testPos, 3, 0, "Y"); break;
		case 11: e = new TxCommitEvent(testPos, 3, 0); break;
		default: e = null; break;
		}
		testPos++;
		return e;
	}


	// SI-Detection-part1.pdf page 62
	public Event testWriteSkewRevisited() {
		Event e = null;
		// H: R1(X) R1(Y) R2(X) R2(Y) W1(X) C1 W2(Y) C2
		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxReadEvent(testPos, 1, 0, "X"); break;
		case 2: e = new TxReadEvent(testPos, 1, 0, "Y"); break;
		
		case 3: e = new TxStartEvent(testPos, 2, 0); break;
		case 4: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		case 5: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		
		case 6: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 7: e = new TxCommitEvent(testPos, 1, 0); break;

		case 8: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 9: e = new TxCommitEvent(testPos, 2, 0); break;
		default: e = null; break;
		}
		testPos++;
		return e;
	}
	
	
	public void createEdges() {
		createReadWrite();
		createWriteWrite();
		createWriteRead();
	}
	
	public void createWriteRead() {

		Iterator<Entry<String, MemCellAccess>> itMain;
		Entry<String, MemCellAccess> entry;
		String memCell;
		MemCellAccess mA;
		
		itMain = memCels.entrySet().iterator();

		while (itMain.hasNext()){
			
			entry = itMain.next();
			memCell = entry.getKey();
			mA = entry.getValue(); 
			
			int writesSize = mA.writes.size();

			for (int i=0 ;i<writesSize; i++){ 
				
				ThreadInfo write = mA.writes.get(i);
				
				long nextWrite;
				
				if ((i+1)<writesSize)
					nextWrite = mA.writes.get(i+1).end;
				else nextWrite = Long.MAX_VALUE;
				
				for (ThreadInfo read : mA.reads) {
				
					if (read.start>write.end && read.start<nextWrite) {
						Edge ed = new Edge(write.transNum, read.transNum,
								WRITE_READ, memCell);
						edges.add(ed);
					}
				}
			}
	
		}
		
		return;
	}
	
	public void createWriteWrite() {

		Iterator<Entry<String, MemCellAccess>> itMain;
		Entry<String, MemCellAccess> entry;
		String memCell;
		MemCellAccess mA;
		
		itMain = memCels.entrySet().iterator();
		
		while (itMain.hasNext()){
			entry = itMain.next();
			memCell = entry.getKey();
			mA = entry.getValue(); 

			int writesSize = mA.writes.size();
			
			for (int i=0 ;i<writesSize; i++){ 
				ThreadInfo a = mA.writes.get(i);
				if ((i+1)==writesSize)
					break;
				
				ThreadInfo b = mA.writes.get(i+1);
				
				Edge ed = new Edge(a.transNum, b.transNum,
						WRITE_WRITE, memCell);
				edges.add(ed);
			}

		}
		
		return;
	}
	
	public void createReadWrite() {

		Iterator<Entry<String, MemCellAccess>> itMain;
		Entry<String, MemCellAccess> entry;
		String memCell;
		MemCellAccess mA;
		
		long start; long end;
		
		itMain = memCels.entrySet().iterator();
		
		while (itMain.hasNext()){
			entry = itMain.next();
			memCell = entry.getKey();
			mA = entry.getValue(); 
			
			for (ThreadInfo write : mA.writes) {

				for (ThreadInfo read : mA.reads) {
					
					start = Math.max(write.start, read.start);
					end = Math.min(write.end, read.end);
					
					if (start < end && write.transNum != read.transNum) {
						Edge ed = new Edge(read.transNum, write.transNum,
								READ_WRITE, memCell);
						edges.add(ed);
					}
				}
			}
	
		}
		
		return;
	}
	
	public void collectInfo() {
		Event evnt;

		while((evnt = reader.getNextEvent()) != null) {
//		while((evnt = test()) != null) {
			ThreadInfo thisThdInfo = thread_trans.get(evnt.threadId);

			if (evnt instanceof TxStartEvent || thisThdInfo == null) {
				t_count++;
				
				thisThdInfo =
					new ThreadInfo(t_count, evnt.timestamp, Long.MAX_VALUE);

				// thread ts.threadId is running transaction number t_count
				thread_trans.put(evnt.threadId, thisThdInfo);
			}
			
			if (evnt instanceof TxReadEvent) {
				thisThdInfo.events.add(evnt);
			}
			else if (evnt instanceof TxWriteEvent) {
				thisThdInfo.events.add(evnt);				
			}
			else if (evnt instanceof TxCommitEvent) {
				thisThdInfo.end = evnt.timestamp;
				
				for (Event e: thisThdInfo.events) {
					if (e instanceof TxReadEvent) {
						TxReadEvent read = (TxReadEvent)e;
						MemCellAccess mA = getMemCellAccess(read.memCel);
						mA.add(thisThdInfo, false); 
					}
					else if (e instanceof TxWriteEvent) {
						TxWriteEvent write = (TxWriteEvent)e;
						MemCellAccess mA = getMemCellAccess(write.memCel);
						mA.add(thisThdInfo, true); 
					}
				}
			}
		}
		return;
	}
	
	private MemCellAccess getMemCellAccess(String memCel) {
		MemCellAccess mA = memCels.get(memCel);
		if (mA == null) {
			mA = new MemCellAccess();
			memCels.put(memCel, mA);
		}
		return mA;
	}

	private void dumpDOT(ArrayList<Edge> edges, int trans_per_dot, String prefix, File outFolder) {

		int count = 0;
		count = 0;

		while (count*trans_per_dot < edges.size()) {
			//			if (count>50)
			//				break;
			try {
				String f = outFolder.getAbsolutePath()+"/"+prefix+count+".dot";
				FileWriter outFile = new FileWriter(f);
				PrintWriter out = new PrintWriter(outFile);

				out.println("digraph \"H-History" + count + "\" {");
				out.println("rankdir=LR");
				out.println();

				Edge ed ;
				HashSet<Integer> usedVert = new HashSet<Integer>();
				int min = Math.min((count+1)*trans_per_dot,edges.size());

				StringBuilder edgString = new StringBuilder();
				for (int i=count*trans_per_dot;i<min;i++) {
					ed = edges.get(i);
					edgString.append("T"+ed.vertA+" -> T"+ed.vertB+" ["+'\n');
					edgString.append("label = \""+ed.label()+"\" ,"+'\n');

					String color;
					switch (ed.type) {
					case WRITE_READ:  color = "black"; break;
					case WRITE_WRITE: color = "green"; break;
					case READ_WRITE:
						edgString.append("style = dashed, \n");
						color = "red"; break;
					default: color = "black"; break;
					}
					edgString.append("color = " + color + "\n"); 
					edgString.append("];"+'\n');
				}

				for (Integer v : usedVert)
					out.println("T"+v);

				out.println(edgString);

				out.println("}");
				out.close();

				count++;

			} catch (IOException e){
				e.printStackTrace();
			}
		}
	}

	private ArrayList<Edge> findWriteSkew() {

		ArrayList<Edge> writeSkew = new ArrayList<Edge>();
		HashMap<Integer, ArrayList<Edge>> readwrites = new HashMap<Integer, ArrayList<Edge>>();

		for (Edge ed : edges) {
			if (ed.type==READ_WRITE) {
				ArrayList<Edge> tmp = readwrites.get(ed.vertA);
				if (tmp==null)
					tmp = new ArrayList<Edge>();
				tmp.add(ed);
				readwrites.put(ed.vertA, tmp);
			}
		}

		System.out.println("writeSkew found: ");
		Iterator<Entry<Integer, ArrayList<Edge>>> it = readwrites.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Integer, ArrayList<Edge>> next = it.next();
			for (Edge ed : next.getValue())
				if (connectBack(next.getKey(), ed.vertB, readwrites))
				{ writeSkew.add(ed); System.out.println("|__ "+ed); }
		}

		return writeSkew;
	}

	private boolean connectBack(int trans1, int trans2, HashMap<Integer, ArrayList<Edge>> readwrites) {
		ArrayList<Edge> edges = readwrites.get(trans2);
		if (edges == null)
			return false;

		for (Edge ed : edges) {
			if (ed.vertB == trans1)
				return true;
		}

		return false;
	}


	// Each thread runs an EventInfo
	// this contains the number of the transaction that the thread is running
	// and the Event info, ReadEvent or WriteEvent
	private class ThreadInfo {
		int transNum;
		long start; // start timestamp
		long end; 	// end timestamp
		ArrayList<Event> events;


		ThreadInfo(int tnum, long s, long e) {
			transNum = tnum;
			start = s;
			end = e;
			events = new ArrayList<Event>();
		}

	}

	private class MemCellAccess {
		ArrayList<ThreadInfo> reads;
		ArrayList<ThreadInfo> writes;
		
		MemCellAccess() {
			reads = new ArrayList<ThreadInfo>();
			writes = new ArrayList<ThreadInfo>();
		}
		void add(ThreadInfo t, boolean w) {
			if (w)
				writes.add(t);
			else reads.add(t);
		}
	}

	private class Edge {
		private int vertA;
		private int vertB;
		private short type; // 0: wr, 1: ww, 2: rw
		private String memCel;

		Edge(int a, int b, short t, String m){
			vertA = a;
			vertB = b;
			type = t;
			memCel = m;
		}
		public String label(){
			switch (type) {
			case WRITE_READ:  return memCel+"-wr";
			case WRITE_WRITE: return memCel+"-ww";
			case READ_WRITE:  return memCel+"-rw";
			default: return "";
			}
		}
		public String toString() {
			return vertA+" -> "+label()+" -> "+vertB;
		}
	}

	public static void main(String[] args) {

		if (args.length < 3) {
			System.out.println("Usage: GenerateDot trace.bin trans_per_dot dest_dir");
			System.out.println("trans_per_dot: Number of transaction per dot file");
			System.exit(1);
		}

		TraceFileReader tfr = new TraceFileReader(new File(args[0]));
		int trans_per_dot = Integer.parseInt(args[1]);
		File dest = new File(args[2]);
		if (!dest.isDirectory()) {
			System.out.println("dest_dir: The pathname must be a directory");
			System.exit(1);
		}

		GenerateDot gd = new GenerateDot(tfr);

		//		gd.processFile();
		gd.dumpDOT(gd.processFile(), trans_per_dot, "history", dest);
		gd.dumpDOT(gd.findWriteSkew(), trans_per_dot, "write-skew", dest);

		tfr.close();

		System.out.println("t_count: " + gd.t_count);
		System.out.println("edges: "   + gd.edges.size());
		System.out.println("memCels: " + gd.memCels.size());
	}

}




/*
 * package pt.unl.fct.di.jtraceview.util;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;

import pt.unl.fct.di.jtraceview.trace.Event;
import pt.unl.fct.di.jtraceview.trace.TraceFileReader;
import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxCommitEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxReadEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxStartEvent;
import pt.unl.fct.di.jtraceview.trace.events.TxWriteEvent;

public class GenerateDot {

	private TraceFileReader reader; // reader of the trace file

	// Types of dependencies
	private static final short WRITE_READ  = 0;
	private static final short WRITE_WRITE = 1;
	private static final short READ_WRITE  = 2;

	// Logs the events on this memCel
	// including the last transaction that wrote and the transactions that read the value
	// HashMap <MemCel, LogEvent>
	private HashMap<String, MemCellAccess> memCels;

	// Thread x is running transactionY
	// HashMap <Thread, Transaction_number>
	private HashMap<Integer,EventInfo> thread_trans;

	// Connections between transactions
	private ArrayList<Edge> edges;

	// Transactions that aborted
	private HashSet<Integer> aborted;

	// Transactions count
	private int t_count;

	private int testPos = 0;

	public GenerateDot(TraceFileReader tfr) {
		reader = tfr;
	}

	public ArrayList<Edge> processFile() {
		// Let's see what transactions abort so we can ignore than later
		clearMem(true);
		collectAbort();

		// Now let's create the dependency edges between transactions
		clearMem(false);
		return createEdges();
	}

	private void clearMem(boolean alsoClearAbort) {
		reader.reset();
		memCels 	 = new HashMap<String,MemCellAccess>();
		thread_trans = new HashMap<Integer,EventInfo>();
		edges		 = new ArrayList<Edge>();
		t_count 	 = 0;
		if (alsoClearAbort)
			aborted  = new HashSet<Integer>();
	}

	// List the transactions that aborted
	private void collectAbort() {
		Event ev;
		while((ev = test()) != null) {
			//while((ev = reader.getNextEvent()) != null) {
			EventInfo thisEvInfo = thread_trans.get(ev.threadId);

			if (ev instanceof TxStartEvent || thisEvInfo == null) {
				t_count++;

				thisEvInfo = new EventInfo (t_count, 0);

				// thread ts.threadId is running transaction number t_count
				thread_trans.put(ev.threadId, thisEvInfo);
			}
			else if (ev instanceof TxAbortEvent) {

				int transNum = thisEvInfo.transNum;
				aborted.add(transNum);
			}

		}
		testPos = 0;
	}
	public Event test() {

		Event e = null;

		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxReadEvent(testPos, 1, 0, "X"); break;

		case 2: e = new TxStartEvent(testPos, 2, 0); break;
		case 3: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		
		case 4: e = new TxReadEvent(testPos, 1, 0, "Y"); break;
		case 5: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		
		case 6: e = new TxReadEvent(testPos, 1, 0, "X"); break;
		case 7: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 8: e = new TxCommitEvent(testPos, 1, 0); break;
		
		case 9: e = new TxReadEvent(testPos, 2, 0, "Y"); break;
		case 10: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 11: e = new TxCommitEvent(testPos, 2, 0); break;

		default: e = null; break;
		}
		testPos++;
		return e;
	}
	public Event test1() {

		Event e = null;

		// TxStartEvent(timestamp, threadId, 0);
		// TxReadEvent(long timestamp, int threadId, int tid, String memCel)
		// TxWriteEvent(long timestamp, int threadId, int tid, String memCel)
		// TxCommitEvent(long timestamp, int threadId, int tid)

		switch (testPos) {
		case 0: e = new TxStartEvent(testPos, 1, 0); break;
		case 1: e = new TxWriteEvent(testPos, 1, 0, "X"); break;
		case 2: e = new TxWriteEvent(testPos, 1, 0, "Y"); break;
		case 3: e = new TxWriteEvent(testPos, 1, 0, "Z"); break;
		case 4: e = new TxCommitEvent(testPos, 1, 0); break;

		case 5: e = new TxStartEvent(testPos, 3, 0); break;
		case 6: e = new TxWriteEvent(testPos, 3, 0, "X"); break;

		case 7: e = new TxStartEvent(testPos, 2, 0); break;
		case 8: e = new TxReadEvent(testPos, 2, 0, "X"); break;
		case 9: e = new TxWriteEvent(testPos, 2, 0, "Y"); break;
		case 10: e = new TxCommitEvent(testPos, 2, 0); break;

		case 11: e = new TxReadEvent(testPos, 3, 0, "Y"); break;
		case 12: e = new TxCommitEvent(testPos, 3, 0); break;
		default: e = null; break;
		}
		testPos++;
		return e;
	}
	public ArrayList<Edge> createEdges() {
		Event ev;

		//while((ev = reader.getNextEvent()) != null) {
		while((ev = test()) != null) {

			EventInfo thisEvInfo = thread_trans.get(ev.threadId);

			if (ev instanceof TxStartEvent || thisEvInfo == null) {
				System.out.println("[thread:"+ev.threadId+"] "+"start");

				t_count++;

				thisEvInfo = new EventInfo (t_count, ev.timestamp);

				// thread ts.threadId is running transaction number t_count
				thread_trans.put(ev.threadId, thisEvInfo);
			}

			// Se é uma transacção que abortou, ignora-a
			if (aborted.contains(thisEvInfo.transNum))
				continue;

			if (ev instanceof TxReadEvent) {
				thisEvInfo.setChangeEv(ev);
			}
			else if (ev instanceof TxWriteEvent) {
				thisEvInfo.setChangeEv(ev);				
			}
			else if (ev instanceof TxCommitEvent) {

				MemCellAccess memCA;
				EventInfo lastWrite;
				Edge ed;

				thisEvInfo.setEnd(ev.timestamp);

				for (Event changeEv : thisEvInfo.changeEv) {

					if (changeEv instanceof TxReadEvent) {
						TxReadEvent thisEv = (TxReadEvent)changeEv;
						System.out.println("[thread:"+ev.threadId+"] "+"read: "+thisEv.memCel);
						// get the mem Access info of this memCel
						memCA = memCels.get(thisEv.memCel);

						// this means that no one has written to this memCel yet
						if (memCA == null) {
							memCA = new MemCellAccess(null);
							memCels.put(thisEv.memCel, memCA);
						}

						// add to the readers list
						memCA.readers.add(thisEvInfo);

						lastWrite = memCA.lastWrite;
						if (lastWrite==null)
							continue;
						
						
						long start = Math.max(lastWrite.start, thisEvInfo.start);
						long end = Math.min(lastWrite.end, thisEvInfo.end);

						// sanity check
						if (start < end
								&& lastWrite.changeEv.get(0).threadId != ev.threadId) {

							// WriteRead dependency between old and this one, over tr.memCel
							ed = new Edge(lastWrite.transNum, thisEvInfo.transNum,
									WRITE_READ, thisEv.memCel);
							edges.add(ed);
						}

					}
					else if (changeEv instanceof TxWriteEvent) {
						TxWriteEvent thisEv = (TxWriteEvent)changeEv;
						System.out.println("[thread:"+ev.threadId+"] "+"write: "+thisEv.memCel);
						MemCellAccess newMemCA = new MemCellAccess(thisEvInfo);

						memCA = memCels.put(thisEv.memCel, newMemCA);

						// this memCell never had a read or write. so no edges to be created for now
						if (memCA == null)
							continue;

						long start; long end;
						
						for (EventInfo read: memCA.readers) {
							start = Math.max(read.start, thisEvInfo.start);
							end = Math.min(read.end, thisEvInfo.end);
							if (start < end && read.changeEv.get(0).threadId != ev.threadId) {
								ed = new Edge(read.transNum, thisEvInfo.transNum,
										READ_WRITE, thisEv.memCel);
								edges.add(ed);
								System.out.println("we got a read write");	
							}
						}

						lastWrite = memCA.lastWrite;
						if (lastWrite==null)
							continue;

						start = Math.max(lastWrite.start, thisEvInfo.start);
						end = Math.min(lastWrite.end, thisEvInfo.end);
						
						// sanity check
						if (start < end
								&& lastWrite.changeEv.get(0).threadId != ev.threadId) {
							// WriteWrite dependency between old and this one, over tr.memCel
							ed = new Edge(lastWrite.transNum, thisEvInfo.transNum,
									WRITE_WRITE, thisEv.memCel);
							edges.add(ed);
						}
					}
				}
			}
		}
		return edges;
	}

	private void dumpDOT(ArrayList<Edge> edges, int trans_per_dot, String prefix, File outFolder) {

		int count = 0;
		count = 0;

		while (count*trans_per_dot < edges.size()) {
			//			if (count>50)
			//				break;
			try {
				String f = outFolder.getAbsolutePath()+"/"+prefix+count+".dot";
				FileWriter outFile = new FileWriter(f);
				PrintWriter out = new PrintWriter(outFile);

				out.println("digraph \"H-History" + count + "\" {");
				out.println("rankdir=LR");
				out.println();

				Edge ed ;
				HashSet<Integer> usedVert = new HashSet<Integer>();
				int min = Math.min((count+1)*trans_per_dot,edges.size());

				StringBuilder edgString = new StringBuilder();
				for (int i=count*trans_per_dot;i<min;i++) {
					ed = edges.get(i);
					edgString.append("T"+ed.vertA+" -> T"+ed.vertB+" ["+'\n');
					edgString.append("label = \""+ed.label()+"\" ,"+'\n');

					String color;
					switch (ed.type) {
					case WRITE_READ:  color = "black"; break;
					case WRITE_WRITE: color = "green"; break;
					case READ_WRITE:
						edgString.append("style = dashed, \n");
						color = "red"; break;
					default: color = "black"; break;
					}
					edgString.append("color = " + color + "\n"); 
					edgString.append("];"+'\n');
				}

				for (Integer v : usedVert)
					out.println("T"+v);

				out.println(edgString);

				out.println("}");
				out.close();

				count++;

			} catch (IOException e){
				e.printStackTrace();
			}
		}
	}

	private ArrayList<Edge> findWriteSkew() {

		ArrayList<Edge> writeSkew = new ArrayList<Edge>();
		HashMap<Integer, ArrayList<Edge>> readwrites = new HashMap<Integer, ArrayList<Edge>>();

		for (Edge ed : edges) {
			if (ed.type==READ_WRITE) {
				ArrayList<Edge> tmp = readwrites.get(ed.vertA);
				if (tmp==null)
					tmp = new ArrayList<Edge>();
				tmp.add(ed);
				readwrites.put(ed.vertA, tmp);
			}
		}

		System.out.println("writeSkew found: ");
		Iterator<Entry<Integer, ArrayList<Edge>>> it = readwrites.entrySet().iterator();
		while (it.hasNext()) {
			Entry<Integer, ArrayList<Edge>> next = it.next();
			for (Edge ed : next.getValue())
				if (connectBack(next.getKey(), ed.vertB, readwrites))
				{ writeSkew.add(ed); System.out.println("|__ "+ed); }
		}

		return writeSkew;
	}

	private boolean connectBack(int trans1, int trans2, HashMap<Integer, ArrayList<Edge>> readwrites) {
		ArrayList<Edge> edges = readwrites.get(trans2);
		if (edges == null)
			return false;

		for (Edge ed : edges) {
			if (ed.vertB == trans1)
				return true;
		}

		return false;
	}


	// Each thread runs an EventInfo
	// this contains the number of the transaction that the thread is running
	// and the Event info, ReadEvent or WriteEvent
	private class EventInfo {
		int transNum;
		long start; // start timestamp
		ArrayList<Event> changeEv;
		long end; 	// end timestamp


		EventInfo(int t, long s) {
			transNum = t;
			start = s;
			end = Long.MAX_VALUE;
			changeEv = new ArrayList<Event>();
		}
		void setChangeEv(Event cEv) {
			changeEv.add(cEv);
		}
		void setEnd(long e) {
			end = e;
		}

	}

	private class MemCellAccess {
		EventInfo lastWrite; // last transaction number that have written this memCel
		HashSet<EventInfo> readers; // list of the transactions that read the value since the last write

		MemCellAccess(EventInfo l) {
			lastWrite = l;
			readers = new HashSet<EventInfo>();
		}
	}

	private class Edge {
		private int vertA;
		private int vertB;
		private short type; // 0: wr, 1: ww, 2: rw
		private String memCel;

		Edge(int a, int b, short t, String m){
			vertA = a;
			vertB = b;
			type = t;
			memCel = m;
		}
		public String label(){
			switch (type) {
			case WRITE_READ:  return memCel+"-wr";
			case WRITE_WRITE: return memCel+"-ww";
			case READ_WRITE:  return memCel+"-rw";
			default: return "";
			}
		}
		public String toString() {
			return vertA+" -> "+label()+" -> "+vertB;
		}
	}

	public static void main(String[] args) {

		if (args.length < 3) {
			System.out.println("Usage: GenerateDot trace.bin trans_per_dot dest_dir");
			System.out.println("trans_per_dot: Number of transaction per dot file");
			System.exit(1);
		}

		TraceFileReader tfr = new TraceFileReader(new File(args[0]));
		int trans_per_dot = Integer.parseInt(args[1]);
		File dest = new File(args[2]);
		if (!dest.isDirectory()) {
			System.out.println("dest_dir: The pathname must be a directory");
			System.exit(1);
		}

		GenerateDot gd = new GenerateDot(tfr);

		//		gd.processFile();
		gd.dumpDOT(gd.processFile(), trans_per_dot, "history", dest);
		gd.dumpDOT(gd.findWriteSkew(), trans_per_dot, "write-skew", dest);

		tfr.close();

		System.out.println("aborted: " + gd.aborted.size());
		System.out.println("t_count: " + gd.t_count);
		System.out.println("edges: "   + gd.edges.size());
		System.out.println("memCels: " + gd.memCels.size());
	}

}


/*
 * 		while((ev = reader.getNextEvent()) != null) {

			if (ev instanceof TxStartEvent) {
				TxStartEvent ts = (TxStartEvent)ev;
				t_count++;

				// thread ts.threadId is running transaction number t_count
				thread_trans.put(ts.threadId, t_count);
			}
			else if (ev instanceof TxReadEvent) {
				TxReadEvent tr = (TxReadEvent)ev;
				Integer tmp = thread_trans.get(tr.threadId);
				if (tmp == null)
					continue;
				int trans_num = tmp.intValue();

				// if this transaction will get aborted later on, we just skip it
				if (aborted.contains(trans_num))
					continue;

				LogEvent le = memCels.get(tr.memCel);
				// create logEvent is this memCel has not been accessed before
				if (le == null) {
					le = new LogEvent(-1);
					memCels.put(tr.memCel, le);
				}
				// Put this trans_num in the readers list of the tr.memCel
				le.readers.add(trans_num);

				Integer old = le.lastWrite;
				// if old == null,no one has written this memCel before,
				// therefore there is no WriteRead dependency
				if (old == null || old == trans_num || old == -1)
					continue;

				// WriteRead dependency between old and this one, over tr.memCel
				Edge ed = new Edge(old, trans_num, WRITE_READ, tr.memCel);
				edges.add(ed);
			}
			else if (ev instanceof TxWriteEvent) {
				TxWriteEvent tw = (TxWriteEvent)ev;
				Integer tmp = thread_trans.get(tw.threadId);
				if (tmp == null)
					continue;
				int trans_num = tmp.intValue();

				// if this transaction will get aborted later on, we just skip it
				if (aborted.contains(trans_num))
					continue;

				LogEvent le =  new LogEvent(trans_num);
				LogEvent old = memCels.put(tw.memCel, le);

				// ReadWrite dependency between old.readers and this transaction, over tr.memCel
				if (old!=null)
					for (Integer r : old.readers) {
						if (r == trans_num) continue;
						Edge ed = new Edge(r, trans_num, READ_WRITE, tw.memCel);
						edges.add(ed);
					}

				// if old == null, no one has written this memCel before,
				// therefore there is no WriteWrite dependency
				if (old == null || old.lastWrite == trans_num || old.lastWrite == -1)
					continue;

				// WriteWrite dependency between old and this one, over tr.memCel
				Edge ed = new Edge(old.lastWrite, trans_num, WRITE_WRITE, tw.memCel);
				edges.add(ed);
			}
		}

 */

