/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.view.tools.component;

import java.awt.Font;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.List;

import javax.swing.JComponent;
import javax.swing.UIDefaults;
import javax.swing.UIManager;

import pt.unl.fct.di.jtraceview.plugins.sequences.FindConflictTransaction;
import pt.unl.fct.di.jtraceview.trace.Event;
import pt.unl.fct.di.jtraceview.trace.TraceFileReader;
import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent;
import pt.unl.fct.di.jtraceview.view.MainForm;
import pt.unl.fct.di.jtraceview.view.tools.component.JThreadSequenceModel.TimeSlice;
import pt.unl.fct.di.jtraceview.view.tools.component.ui.BasicJThreadSequenceUI;
import pt.unl.fct.di.jtraceview.view.tools.component.ui.JThreadSequenceUI;

@SuppressWarnings("serial")
public class JThreadSequence extends JComponent {
	
	/**
	 * The UI class ID string.
	 */
	private static final String uiClassID = "JThreadSequenceUI";
	
	/**
	 * Sets the new UI delegate.
	 * 
	 * @param ui
	 *            New UI delegate.
	 */
	public void setUI(JThreadSequenceUI ui) {
		super.setUI(ui);
	}


	/**
	 * Resets the UI property to a value from the current look and feel.
	 * 
	 * @see JComponent#updateUI
	 */
	public void updateUI() {
		if (UIManager.get(getUIClassID()) != null) {
			setUI((JThreadSequenceUI) UIManager.getUI(this));
		} else {
			setUI(new BasicJThreadSequenceUI());
		}
	}


	/**
	 * Returns the UI object which implements the L&F for this component.
	 * 
	 * @return UI object which implements the L&F for this component.
	 * @see #setUI
	 */
	public JThreadSequenceUI getUI() {
		return (JThreadSequenceUI) ui;
	}


	/**
	 * Returns the name of the UI class that implements the L&F for this
	 * component.
	 * 
	 * @return The name of the UI class that implements the L&F for this
	 *         component.
	 * @see JComponent#getUIClassID
	 * @see UIDefaults#getUI
	 */
	public String getUIClassID() {
		return uiClassID;
	}

	protected JThreadSequenceModel model;
	protected int threadHeight;
	protected int threadGap;
	protected double timeSliceWidthFactor;
	
	protected Font titleFont;
	protected Font yLabelFont;
	protected Font xLabelFont;
	protected Font sliceLabelFont;
	protected String title;
	protected boolean buttonSelected = false;
	protected boolean buttonColor = true;
	
	public JThreadSequence() {
		this("Test Thread Sequence", new DefaultJThreadSequenceModel());
	}
	
	public JThreadSequence(String title, JThreadSequenceModel model) {
		this.model = model;
		this.threadHeight = 40;
		this.threadGap = 20;
		this.timeSliceWidthFactor = 0.0001;
		this.title = title;
		
		titleFont = new Font(Font.SANS_SERIF, Font.PLAIN, 30);
		yLabelFont = new Font(Font.SANS_SERIF, Font.PLAIN, 15);
		sliceLabelFont = new Font(Font.SANS_SERIF, Font.PLAIN, 15);
		xLabelFont = new Font(Font.SANS_SERIF, Font.PLAIN, 10);
		
		this.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(MouseEvent evt) {
				TimeSlice ts = ((BasicJThreadSequenceUI)getUI()).selectSlice(evt.getX(), evt.getY());
				if (ts != null) {
					//TimeSlice conf = FindConflictTransaction.findConflict(ts, MainForm.getMainForm().getTraceFile());
					if (ts.userData.get("last.event") instanceof TxAbortEvent) {
						TxAbortEvent abt = (TxAbortEvent)ts.userData.get("last.event");
						long s = (Long)ts.userData.get("savepoint");
						TraceFileReader reader = new TraceFileReader(MainForm.getMainForm().getTraceFile());
						List<Integer> threads = FindConflictTransaction.getThreads(reader);
						reader.restoreSavePoint(s);
						Event conf = FindConflictTransaction.getConflictTransaction(abt, threads, reader);
						if (conf != null) {
							Long offset = (Long)ts.userData.get("offset");
							((BasicJThreadSequenceUI)getUI()).selectSlice(conf.timestamp-offset);
							
						}
						reader.close();
					}
					JThreadSequence.this.repaint();
				}
			}
		});
		
		this.updateUI();
	}
	
	public JThreadSequenceModel getModel() {
		return model;
	}
	
	public void setModel(JThreadSequenceModel model) {
		this.model = model;
		this.updateUI();
		this.validate();
	}
	
	

	public Font getSliceLabelFont() {
		return sliceLabelFont;
	}


	public void setSliceLabelFont(Font sliceLabelFont) {
		this.sliceLabelFont = sliceLabelFont;
	}


	public int getThreadHeight() {
		return threadHeight;
	}
	
	public void setThreadHeight(int height) {
		this.threadHeight = height;
	}

	public int getThreadGap() {
		return threadGap;
	}

	public void setThreadGap(int threadGap) {
		this.threadGap = threadGap;
	}


	public double getTimeSliceWidthFactor() {
		return timeSliceWidthFactor;
	}


	public void setTimeSliceWidthFactor(double timeSliceWidthFactor) {
		this.timeSliceWidthFactor = timeSliceWidthFactor;
	}


	public Font getTitleFont() {
		return titleFont;
	}


	public void setTitleFont(Font titleFont) {
		this.titleFont = titleFont;
	}


	public Font getYLabelFont() {
		return yLabelFont;
	}


	public void setYLabelFont(Font labelFont) {
		yLabelFont = labelFont;
	}


	public Font getXLabelFont() {
		return xLabelFont;
	}


	public void setXLabelFont(Font labelFont) {
		xLabelFont = labelFont;
	}


	public String getTitle() {
		return title;
	}


	public void setTitle(String title) {
		this.title = title;
	}


	public void setSelected(boolean b){
		buttonSelected = b;
	}
	public boolean getSelected(){
		return buttonSelected;
	}
	public void setColor(){
		buttonColor = true;
	}
	public void setBlackWhite(){
		buttonColor = false;
	}
	public boolean getButtonColor(){
		return buttonColor;
	}
	
	
	
	
	

}
