/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.view.tools.component.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Paint;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.TexturePaint;
import java.awt.image.BufferedImage;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.border.EmptyBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.plaf.ComponentUI;

import pt.unl.fct.di.jtraceview.plugins.sequences.FindConflictTransaction;
import pt.unl.fct.di.jtraceview.trace.Event;
import pt.unl.fct.di.jtraceview.trace.TraceFileReader;
import pt.unl.fct.di.jtraceview.trace.events.TxAbortEvent;
import pt.unl.fct.di.jtraceview.util.Pair;
import pt.unl.fct.di.jtraceview.view.MainForm;
import pt.unl.fct.di.jtraceview.view.tools.component.JThreadSequence;
import pt.unl.fct.di.jtraceview.view.tools.component.JThreadSequenceModel;
import pt.unl.fct.di.jtraceview.view.tools.component.JThreadSequenceModel.TimeSlice;

public class BasicJThreadSequenceUI extends JThreadSequenceUI {
	
	protected static final int minimumHeight = 400;
	protected static final int minimumWidth = 400;
	
	protected JThreadSequence threadSequence;
	
	// component building blocks
	protected Rectangle title;
	protected Rectangle yLabel;
	protected Rectangle xLabel;
	protected Rectangle sequence;
	protected Rectangle window;
	protected Rectangle slice;
	
	protected JLabel titleLabel;
	protected JLabel[] threadLabels;
	protected JLabel[] sliceLabels;
	protected long maxXLabel;
	protected HashMap<String,Pair<Color,Color>> sliceColors;
	protected HashMap<String,Integer> slicePatterns;
	protected boolean computed = false;
	protected boolean color = true;
	
	protected Color[] colors = {new Color(0xFF0000),new Color(0x00CC00),new Color(0xCCCCCC),new Color(0x00DDDD),new Color(0xFFFF00), 
			new Color(0xCC33FF),new Color(0xFF9966)};
	//vermelho verde cinzento azul amarelo roxo salmao
	protected Color[] secondColors = {new Color(0x990000),new Color(0x006600),new Color(0x777777),new Color(0x0066CC),new Color(0xFF9900),
			new Color(0x9900CC),new Color(0xFF3366)};
	
	private ChangeListener jThreadSequenceChangeListener;
	

	private List<SliceSelected> selectedList =  new LinkedList<SliceSelected>();
	private List<SliceSelected> joinList = new LinkedList<SliceSelected>();
	
	public static ComponentUI createUI(JComponent c) {
		return new BasicJThreadSequenceUI();
	}
	
	public void installUI(JComponent c) {
		this.threadSequence = (JThreadSequence)c;
		installDefaults();
		installComponents();
		installListeners();

		c.setLayout(createLayoutManager());
		c.setBorder(new EmptyBorder(10, 10, 10, 10));
		
		calculateComponentBounds();
		
		
	}
	
	public void installComponents() {
		titleLabel = new JLabel(threadSequence.getTitle(), JLabel.CENTER);
		titleLabel.setFont(threadSequence.getTitleFont());
		
		titleLabel.setBounds(0, 0, titleLabel.getPreferredSize().width, 
				titleLabel.getPreferredSize().height);
		threadSequence.add(titleLabel);
		
		JThreadSequenceModel model = threadSequence.getModel();
		threadLabels = new JLabel[model.getThreadCount()];
		int i=0;
		
		sliceColors = new HashMap<String,Pair<Color,Color>>();
		slicePatterns = new HashMap<String,Integer>();
		
		for (JThreadSequenceModel.Thread t : model.getThreads()) {
			threadLabels[i] = new JLabel(t.getId(), JLabel.CENTER);
			threadLabels[i].setFont(threadSequence.getYLabelFont());
			threadSequence.add(threadLabels[i++]);
		
			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
				sliceColors.put(ts.id, null);
				slicePatterns.put(ts.id, null);
			}
		}
		
		sliceLabels = new JLabel[sliceColors.size()];
		i=0;
		for (Map.Entry<String,Pair<Color,Color>> e : sliceColors.entrySet()) {
			sliceLabels[i] = new JLabel(e.getKey(), JLabel.CENTER);
			sliceLabels[i].setFont(threadSequence.getSliceLabelFont());
			e.setValue(new Pair<Color,Color>(colors[i%colors.length],secondColors[i%colors.length]));
			sliceLabels[i].setOpaque(true);
			sliceLabels[i].setBackground(colors[i%colors.length]);
			threadSequence.add(sliceLabels[i++]);
		}
		i=0;
		for (Map.Entry<String,Integer> e : slicePatterns.entrySet()) {
			e.setValue(i%5);
			i++;
		}	
	}
	
	private void calculateComponentBounds() {
		Insets ins = threadSequence.getInsets();
		
		title = new Rectangle(0, 0, titleLabel.getPreferredSize().width, 
				titleLabel.getPreferredSize().height);
		
		
		int width = 0;
		int height = 10;
		
		JThreadSequenceModel model = threadSequence.getModel();
		List<JThreadSequenceModel.Thread> threads = model.getThreads();
		
		for (JThreadSequenceModel.Thread t : threads) {
			height += threadSequence.getThreadGap()+threadSequence.getThreadHeight();
			
//			JThreadSequenceModel.TimeSlice prev = null;
//			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
//				if (prev != null) {
//					width += (int)((ts.start-prev.end)*threadSequence.getTimeSliceWidthFactor());
//				}
//				width += (int)((ts.end-ts.start)*threadSequence.getTimeSliceWidthFactor());
//				prev = ts;
//			}
			
			if ((int)(t.getSlices().get(t.getSlices().size()-1).end*threadSequence.getTimeSliceWidthFactor()) > width)
				width = (int)(t.getSlices().get(t.getSlices().size()-1).end*threadSequence.getTimeSliceWidthFactor());
			
		}
		
		height += 10;
		
		if (height < minimumHeight) {
			height = minimumHeight;
		}
		
		if (width == 0) {
			width = minimumWidth;
		}
		
		int maxWidth = 0;
		for (int i=0; i < threadLabels.length; i++) {
			if (threadLabels[i].getPreferredSize().width > maxWidth)
				maxWidth = threadLabels[i].getPreferredSize().width;
		}
		
		sequence = new Rectangle(0, 0, width, height);
		
		yLabel = new Rectangle(0, 0, maxWidth, height);
		
		JLabel t = new JLabel("M", JLabel.CENTER);
		t.setFont(threadSequence.getXLabelFont());
		
		xLabel = new Rectangle(0, 0, width, t.getPreferredSize().height);
		
		int sliceWidth = 0;
		int maxHeight = 0;
		for (int i=0; i < sliceLabels.length; i++) {
			sliceWidth += 20+sliceLabels[i].getPreferredSize().width;
			if (sliceLabels[i].getPreferredSize().height > maxHeight)
				maxHeight = sliceLabels[i].getPreferredSize().height;
		}
		
		slice = new Rectangle(0, 0, sliceWidth, maxHeight);
		
		window = new Rectangle(0, 0, ins.left+10+yLabel.width+10+sequence.width+10+ins.right,
				ins.top+10+title.height+10+sequence.height+5+xLabel.height+10+ins.bottom+
				20+slice.height);
		
	}

	protected LayoutManager createLayoutManager() {
		return new JThreadSequenceLayout();
	}

	public void uninstallUI(JComponent c) {
		c.setLayout(null);
		uninstallListeners();
		uninstallComponents();
		uninstallDefaults();
		
		this.threadSequence.removeAll();

		this.threadSequence = null;
		this.titleLabel = null;
	}
	
	public void uninstallDefaults() {
	}

	public void uninstallComponents() {
	}

	public void uninstallListeners() {
	}

	public void installListeners() {
		this.jThreadSequenceChangeListener = new ChangeListener() {
			public void stateChanged(ChangeEvent e) {
				threadSequence.validate();
				threadSequence.repaint();
			}

		};

		this.threadSequence.getModel().addChangeListener(
				this.jThreadSequenceChangeListener);
	}

	

	public void installDefaults() {		
	}

	@Override
	public void paint(Graphics g, JComponent c) {
		super.paint(g, c);
		
		drawModel((Graphics2D)g);
		
	}
	
	private void drawModel(Graphics2D g) {
		
		g.setColor(Color.BLACK);
		
		//drawAxis(g);
		//g.draw(title);
		//g.draw(yLabel);
		//g.draw(sequence);
		drawAxis(g);
		//drawXLabel(g);
		drawThreadTimeSlice(g);
		drawSelectedSlice(g);
		color = threadSequence.getButtonColor();
		if(threadSequence.getSelected()){
			if(!computed)
				computeAllAborts();
			computed = true;
			drawAllSelectedSlices(g);
		}
	}
	
	private void drawThreadTimeSlice(Graphics2D g) {
		JThreadSequenceModel model = threadSequence.getModel();
		
		int y = sequence.y;
		int dy = sequence.height/(model.getThreadCount()+1);
		
		
		for (JThreadSequenceModel.Thread t : model.getThreads()) {
			y += dy;
			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
				int x = sequence.x+(int)(ts.start*threadSequence.getTimeSliceWidthFactor());
				int width = (int)((ts.end-ts.start)*threadSequence.getTimeSliceWidthFactor());
				Paint p = new GradientPaint(x, y-(threadSequence.getThreadHeight()/2),
						sliceColors.get(ts.id).first,x+width,
						y-(threadSequence.getThreadHeight()/2)+threadSequence.getThreadHeight(),
						sliceColors.get(ts.id).second);
				if(color)
					g.setPaint(p);
				else{
					g.setPaint(getTextureImage(slicePatterns.get(ts.id)));
				}
				g.fillRect(x, y-(threadSequence.getThreadHeight()/2),width, threadSequence.getThreadHeight());

				g.setColor(Color.BLACK);
				g.drawRect(x, y-(threadSequence.getThreadHeight()/2),width, threadSequence.getThreadHeight());
			}
		}
		
	}
	
	public TimeSlice selectSlice(long ts_start) {
		JThreadSequenceModel model = threadSequence.getModel();	
		int y = sequence.y;
		int dy = sequence.height/(model.getThreadCount()+1);
		
		
		for (JThreadSequenceModel.Thread t : model.getThreads()) {
			y += dy;
			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
				int x = sequence.x+(int)(ts.start*threadSequence.getTimeSliceWidthFactor());
				int width = (int)((ts.end-ts.start)*threadSequence.getTimeSliceWidthFactor());
				Rectangle r = new Rectangle(x, y-(threadSequence.getThreadHeight()/2), 
						width, threadSequence.getThreadHeight());

				if (ts.start == ts_start) {
//					if (selected == null) {
//						selected = new SliceSelected(ts, r);
//					}
//					else if (selected != null && join == null) {
//						join = new SliceSelected(ts, r);
//					}
//					else {
//						join = null;
//						selected = new SliceSelected(ts, r);
//					}
					join = new SliceSelected(ts, r);
					return ts;
				}
			}
		}
		return null;
	}
	
	public TimeSlice selectSlice(int xD, int yD) {
		JThreadSequenceModel model = threadSequence.getModel();	
		int y = sequence.y;
		int dy = sequence.height/(model.getThreadCount()+1);
		
		for (JThreadSequenceModel.Thread t : model.getThreads()) {
			y += dy;
			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
				
				//g.setColor(getTidColor(tids, ts));
				int x = sequence.x+(int)(ts.start*threadSequence.getTimeSliceWidthFactor());
				int width = (int)((ts.end-ts.start)*threadSequence.getTimeSliceWidthFactor());
				Rectangle r = new Rectangle(x, y - (threadSequence.getThreadHeight()/2), 
						width, threadSequence.getThreadHeight());

				if (r.contains(xD, yD)) {
	//				System.out.println("Hit Box: "+ts);
//					if (selected == null) {
//						selected = new SliceSelected(ts, r);
//					}
//					else if (selected != null && join == null) {
//						join = new SliceSelected(ts, r);
//					}
//					else {
//						join = null;
//						selected = new SliceSelected(ts, r);
//					}
					selected = new SliceSelected(ts, r);
					join = null;
					return ts;
				}
			}
		}
		return null;
	}
	
	private class SliceSelected {
//		public TimeSlice selected;
		public Rectangle rect;
		public SliceSelected(TimeSlice selected, Rectangle rect) {
//			this.selected = selected;
			this.rect = rect;
		}
		
	}
	
	private SliceSelected selected = null;
	private SliceSelected join = null;
	
	private void drawSelectedSlice(Graphics2D g) {
		if (selected != null) {
			g.setColor(Color.BLACK);
			g.drawRect(selected.rect.x, selected.rect.y, 
					selected.rect.width, selected.rect.height);
		}
		if (join != null) {
			g.setColor(Color.BLACK);
			g.drawRect(join.rect.x, join.rect.y, 
					join.rect.width, join.rect.height);
			
			Point p1 = new Point(selected.rect.x + selected.rect.width/2,
					selected.rect.y + selected.rect.height/2);
			Point p2 = new Point(join.rect.x + join.rect.width/2,
					join.rect.y + join.rect.height/2);
			drawArrow(g, p1, p2, 15);
		}
	}
	public void computeAllAborts(){
		JThreadSequenceModel model = threadSequence.getModel();
		threadLabels = new JLabel[model.getThreadCount()];
		int i=0;
		
		TraceFileReader reader = MainForm.getMainForm().getTraceReader();
		List<Integer> threads = FindConflictTransaction.getThreads(reader);
		
		int y = sequence.y;
		int dy = sequence.height/(model.getThreadCount()+1);
		for (JThreadSequenceModel.Thread t : model.getThreads()) {
			threadLabels[i] = new JLabel(t.getId(), JLabel.CENTER);
			threadLabels[i].setFont(threadSequence.getYLabelFont());
			threadSequence.add(threadLabels[i++]);
			y += dy;
			for (JThreadSequenceModel.TimeSlice ts : t.getSlices()) {
				if (ts.userData.get("last.event") instanceof TxAbortEvent) {
					TxAbortEvent abt = (TxAbortEvent)ts.userData.get("last.event");
					long s = (Long)ts.userData.get("savepoint");
//					TraceFileReader reader = new TraceFileReader(MainForm.getMainForm().getTraceFile());
//					reader.restoreSavePoint(s);
//					Event conf = FindConflictTransaction.getConflictTransaction(abt, reader);
//					reader.close();
					//TraceFileReader reader = MainForm.getMainForm().getTraceReader();
					long currentPos = reader.setSavePoint();
					reader.restoreSavePoint(s);
					Event conf = FindConflictTransaction.getConflictTransaction(abt, threads, reader);
					reader.restoreSavePoint(currentPos);
					
					if (conf != null) {
						int x = sequence.x+(int)(ts.start*threadSequence.getTimeSliceWidthFactor());
						int width = (int)((ts.end-ts.start)*threadSequence.getTimeSliceWidthFactor());
						Rectangle r = new Rectangle(x, y-(threadSequence.getThreadHeight()/2), 
								width, threadSequence.getThreadHeight());

						Long offset = (Long)ts.userData.get("offset");
						selectedList.add(new SliceSelected(ts,r));
						selectSlice(conf.timestamp-offset);	
						joinList.add(join);
					}
				}
			}
		}
		selected = null;
		join = null;
	}
	public void drawAllSelectedSlices(Graphics2D g) {
		Iterator<SliceSelected> sIterator = selectedList.iterator();
		Iterator<SliceSelected> jIterator = joinList.iterator();
		
		for(int i=0;i<selectedList.size();i++){
			SliceSelected s = sIterator.next();
			SliceSelected j = jIterator.next();

			//System.out.println("for "+i);
			g.setColor(Color.BLACK);
			
			g.drawRect(s.rect.x, s.rect.y,s.rect.width, s.rect.height);
			g.drawRect(j.rect.x, j.rect.y,j.rect.width, j.rect.height);
			
			Point p1 = new Point(s.rect.x + s.rect.width/2,
					s.rect.y + s.rect.height/2);
			Point p2 = new Point(j.rect.x + j.rect.width/2,
					j.rect.y + j.rect.height/2);
			drawArrow(g, p1, p2, 15);
		}
	}
	private TexturePaint getTextureImage(int num) {	
		int size = 8;
		BufferedImage bi = new BufferedImage(size, size,
				BufferedImage.TYPE_INT_ARGB);
		TexturePaint imagePaint = null; 
		Graphics2D g2 = bi.createGraphics();
		switch(num){
			case 0:	//diagonais
//				int[] x1 = { 0, 0, 2, 6 };
//				int[] y1 = { 2, 6, 8, 8 };
				g2.setPaint(Color.BLACK);
//				g2.fillPolygon(x1, y1, 4);
//				g2.fillPolygon(y1, x1, 4);
				g2.drawLine(0,0,8,8);
				imagePaint = new TexturePaint(bi,new Rectangle(0,0,20,20));
			break;
			case 1://quadradinhos pequenos
				g2.setPaint(Color.BLACK);
				g2.fillRect(3,3,2,2);
				imagePaint = new TexturePaint(bi,new Rectangle(0,0,15,15));
			break;
			case 2://bolas
				g2.setPaint(Color.BLACK);
				g2.drawOval(0,0,8,8);
				imagePaint = new TexturePaint(bi,new Rectangle(0,0,15,15));
			break;
			case 3://losanglos
				int[] x2 = {0,4,8,4};
				int[] y2 = {4,8,4,0};
				g2.setPaint(Color.BLACK);
				g2.fillPolygon(x2, y2, 4);
				imagePaint = new TexturePaint(bi, new Rectangle(0,0,15,15));
			break;
			default://cruzes
				g2.setPaint(Color.BLACK);
				g2.drawLine(0,0,8,8);
				g2.drawLine(0,8,8,0);
				imagePaint = new TexturePaint(bi, new Rectangle(0,0,20,20));
			break;
		}
		return imagePaint;
	}

//	private void drawXLabel(Graphics2D g) {
//		g.setFont(threadSequence.getXLabelFont());
//		for (long i=0; i < maxXLabel; i++) {
//			g.drawString(""+i, xLabel.x + 
//					(int)(i*threadSequence.getTimeSliceWidthFactor()), xLabel.y+xLabel.height-1);
//		}
//		
//	}

	private void drawAxis(Graphics2D g) {
		Point axisRoot = new Point(sequence.x, sequence.y+sequence.height);
		Point yCoord = new Point(sequence.x, sequence.y);
		Point xCoord = new Point(sequence.x+sequence.width, sequence.y+sequence.height);
		
		drawArrow(g, axisRoot, yCoord, 15);
		drawArrow(g, axisRoot, xCoord, 15);
	}
	
	private void drawArrow(Graphics2D g, Point p1, Point p2, int size) {
		g.drawLine(p1.x, p1.y, p2.x, p2.y);
		double alpha = Math.atan((p2.y-p1.y)/(double)(p2.x-p1.x));
		
		if (/*p2.y > p1.y || */p2.x < p1.x) {
			alpha += Math.PI;
		}
		
		double x1 = size*Math.cos(((5.0/6.0)*Math.PI)+alpha)+p2.x;
		double y1 = size*Math.sin(((5.0/6.0)*Math.PI)+alpha)+p2.y;
		
		double x2 = size*Math.cos(((5.0/6.0)*Math.PI)+(Math.PI/3.0)+alpha)+p2.x;
		double y2 = size*Math.sin(((5.0/6.0)*Math.PI)+(Math.PI/3.0)+alpha)+p2.y;
		
		g.drawLine(p2.x, p2.y, (int)x1, (int)y1);
		g.drawLine(p2.x, p2.y, (int)x2, (int)y2);
	}

	protected class JThreadSequenceLayout implements LayoutManager {
		
		public void addLayoutComponent(String name, Component c) {
		}

		public void removeLayoutComponent(Component c) {
		}

		public Dimension preferredLayoutSize(Container c) {
			return new Dimension(window.width, window.height);
		}

		public Dimension minimumLayoutSize(Container c) {
			return this.preferredLayoutSize(c);
		}

		public void layoutContainer(Container c) {
			Insets ins = threadSequence.getInsets();
			
			title.x = ins.left+10;
			title.y = ins.top+10;
			
			yLabel.x = ins.left+10;
			yLabel.y = title.y+title.height+10;
			
			sequence.x = yLabel.x+yLabel.width+10;
			sequence.y = yLabel.y;
			
			xLabel.x = sequence.x;
			xLabel.y = sequence.y+sequence.height+5;
			
			titleLabel.setBounds(title);
			
			JThreadSequenceModel model = threadSequence.getModel();
			int y = yLabel.y;
			int dy = yLabel.height/(model.getThreadCount()+1);
			
			for (int i=0; i < threadLabels.length; i++) {
				y += dy;
				threadLabels[i].setBounds(yLabel.x, y-(threadSequence.getThreadHeight()/2), 
						yLabel.width, threadSequence.getThreadHeight());
			}
			
			y = xLabel.y;
			int x = xLabel.x;
			
			long max = 0;
			
			for (JThreadSequenceModel.Thread t : model.getThreads()) {
				if (t.getSlices().get(t.getSlices().size()-1).end > max)
					max = t.getSlices().get(t.getSlices().size()-1).end; 
			}
			
			maxXLabel = max;
			
			x = xLabel.x;
			for (JLabel l : sliceLabels) {
				l.setBounds(x, xLabel.y+xLabel.height+20, l.getPreferredSize().width, 
						l.getPreferredSize().height);
				x += 20+l.getPreferredSize().width;
			}
		}

	}
}
