/*  Copyright (C) 2010  Faculdade de Ciências e Tecnologia
 *                      Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package pt.unl.fct.di.jtraceview.view.tree;

import java.awt.Cursor;
import java.util.ArrayList;
import java.util.List;

import javax.swing.SwingWorker;

import org.jfree.chart.JFreeChart;

import pt.unl.fct.di.jtraceview.view.MainForm;

@SuppressWarnings("serial")
public class ChartGroupTreeItem extends GroupTreeItem {

	private static final String title = "Charts";
	
	public ChartGroupTreeItem() {
		super(title);
		
		createChartTools();
	}

	private void createChartTools() {
		this.add(new AbortTreeItem());
		this.add(new CommitAbortTreeItem());
		this.add(new TransactionIdTreeItem());
		this.add(new ReadWriteTransactionTreeItem());
		this.add(new XYChartItem());
		this.add(new AccessMemTreeItem());
		this.add(new TransactionRetryTreeItem());
		this.add(new TransactionLifetimeTreeItem());
		this.add(new AbortReasonTreeItem());
		this.add(new WastedWorkTreeItem());
	}

	@Override
	public void itemSelected() {
		if (MainForm.getMainForm().getTraceFile() == null)
			MainForm.getMainForm().openTraceFile();
		
		MainForm.getMainForm().setMode(MainForm.CHART_MODE);
		
		(new SwingWorker<List<JFreeChart>,Object>() {
			@Override
			protected List<JFreeChart> doInBackground() throws Exception {
				MainForm.getMainForm().getProgressBar().setIndeterminate(true);
				MainForm.getMainForm().setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));

				List<JFreeChart> charts = new ArrayList<JFreeChart>();				
				for (int i=0; i < getChildCount(); i++) {
					MultipleChartTreeItem item = (MultipleChartTreeItem)getChildAt(i);
					charts.addAll(item.getCharts());
				}
				return charts;
			}
			
			@Override
			protected void done() {
				try { 
					MainForm.getMainForm().getChartPanel().setCharts(get());
					MainForm.getMainForm().setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
					MainForm.getMainForm().getProgressBar().setIndeterminate(false);
		        } catch (Exception ignore){}
			}

		}).execute();
		
		
	}

}
