/*  Copyright (C) 2007 Faculdade de Ciências e Tecnologia 
 *                     Universidade Nova de Lisboa
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>
#include <assert.h>
#include "buffer.h"
#include "xalloc.h"

#define min(x,y) (((x) < (y)) ? (x) : (y))

void
buffer_init(struct buffer *buf, size_t size, void *user_data,
	    int (*flush_func)(void *, size_t, void *))
{
	buf->data=xmalloc(size);
	buf->size=0;
	buf->max_size=size;
	buf->user_data=user_data;
	buf->flush_func=flush_func;
}

void 
buffer_free(struct buffer *buf)
{
	if (buf->size)
		buffer_flush(buf);

	assert(!buf->size);

	free(buf->data);
}

int
buffer_flush(struct buffer *buf)
{
	int z;

	z=buf->flush_func(buf->data,buf->size,buf->user_data);
	buf->size=0;

	return z;
}

int
buffer_write(struct buffer *buf, void *data, size_t size)
{
	int to_copy;

	while (size)
	{
		to_copy=min(size,buf->max_size-buf->size);
		memcpy(buf->data+buf->size,data,to_copy);

		size-=to_copy;
		data+=to_copy;

		buf->size+=to_copy;

		assert(size >= 0);
		assert(buf->size <= buf->max_size);

		if (size)
		{
			int z;

			z=buffer_flush(buf);

			if (z)
				return z;
		}
	}

	return 0;
}
